"""Tests for mtg_ssm.serialization.xlsx."""

from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple
from uuid import UUID

import openpyxl
import pytest
from syrupy.assertion import SnapshotAssertion

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.containers.collection import MagicCollection
from mtg_ssm.containers.counts import CountType, ScryfallCardCount
from mtg_ssm.containers.indexes import Oracle
from mtg_ssm.serialization import xlsx


@pytest.fixture(scope="session")
def oracle(scryfall_data: ScryfallDataSet) -> Oracle:
    """Oracle fixture."""
    accepted_sets = {
        "lea",
        "fem",
        "s00",
        "ice",
        "hop",
        "mbs",
        "pmbs",
        "chk",
        "bok",
        "sok",
        "neo",
        "pneo",
    }
    scryfall_data2 = ScryfallDataSet(
        sets=[s for s in scryfall_data.sets if s.code in accepted_sets],
        cards=[c for c in scryfall_data.cards if c.set in accepted_sets],
        migrations=[],
    )
    return Oracle(scryfall_data2)


def test_create_all_sets(snapshot: SnapshotAssertion, oracle: Oracle) -> None:
    book = openpyxl.Workbook()
    sheet = book.create_sheet()
    xlsx.create_all_sets(sheet, oracle.index)
    assert [[cell.value for cell in row] for row in sheet.rows] == snapshot(name=str(sheet.title))


def test_create_haverefs(oracle: Oracle) -> None:
    fem_thallids = [c for c in oracle.index.name_to_cards["Thallid"] if c.set == "fem"]
    fem_thallids.sort(key=lambda c: c.collector_number)
    haverefs = xlsx.create_haverefs(oracle.index, "fem", fem_thallids)
    assert haverefs == "SUM('FEM'!A2:A5)"


@pytest.mark.parametrize(
    ("name", "exclude_sets", "expected"),
    [
        ("Forest", None, None),
        ("Rhox", None, "=IF('S00'!A2>0,\"S00:\"&'S00'!A2,\"\")"),
        ("Rhox", {"s00"}, None),
        (
            "Dark Ritual",
            {"lea"},
            '=_xlfn.TEXTJOIN(", ",1,IF(\'ICE\'!A2>0,"ICE:"&\'ICE\'!A2,""),IF(\'HOP\'!A3>0,"HOP:"&\'HOP\'!A3,""))',
        ),
        ("Dark Ritual", {"lea", "ice"}, "=IF('HOP'!A3>0,\"HOP:\"&'HOP'!A3,\"\")"),
        (
            "Thallid",
            None,
            "=IF(SUM('FEM'!A2:A5)>0,\"FEM:\"&SUM('FEM'!A2:A5),\"\")",
        ),
    ],
)
def test_get_references(
    oracle: Oracle, name: str, exclude_sets: Optional[Set[str]], expected: str
) -> None:
    print_refs = xlsx.get_references(oracle.index, name, exclude_sets=exclude_sets)
    assert print_refs == expected


def test_create_all_cards_sheet(snapshot: SnapshotAssertion, oracle: Oracle) -> None:
    book = openpyxl.Workbook()
    sheet = book.create_sheet()
    xlsx.create_all_cards(sheet, oracle.index)
    assert [[cell.value for cell in row] for row in sheet.rows] == snapshot(name=str(sheet.title))


def test_create_set_sheet(snapshot: SnapshotAssertion, oracle: Oracle) -> None:
    card_counts: ScryfallCardCount = {
        UUID("fbdcbd97-90a9-45ea-94f6-2a1c6faaf965"): {CountType.NONFOIL: 1},
        UUID("b346b784-7bde-49d0-bfa9-56236cbe19d9"): {CountType.FOIL: 2},
        UUID("768c4d8f-5700-4f0a-9ff2-58422aeb1dac"): {
            CountType.NONFOIL: 3,
            CountType.FOIL: 4,
        },
    }
    collection = MagicCollection(oracle=oracle, counts=card_counts)
    book = openpyxl.Workbook()
    sheet = book.create_sheet()
    xlsx.create_set_sheet(sheet, collection, "ice")
    assert [[cell.value for cell in row] for row in sheet.rows] == snapshot(name=str(sheet.title))


def test_write(snapshot: SnapshotAssertion, oracle: Oracle, tmp_path: Path) -> None:
    xlsx_path = tmp_path / "outfile.xlsx"
    card_counts: ScryfallCardCount = {
        UUID("5d5f3f57-410f-4ee2-b93c-f5051a068828"): {
            CountType.NONFOIL: 7,
            CountType.FOIL: 12,
        }
    }
    collection = MagicCollection(oracle=oracle, counts=card_counts)

    serializer = xlsx.XlsxDialect()
    serializer.write(xlsx_path, collection)

    workbook = openpyxl.load_workbook(filename=xlsx_path)
    for i, sheet in enumerate(workbook.worksheets):
        assert [[cell.value for cell in row] for row in sheet.rows] == snapshot(
            name=f"{i:02d} - {sheet.title}"
        )


@pytest.mark.parametrize(
    ("sheets_and_rows", "skip_sheets", "expected"),
    [
        pytest.param(
            [("MySheet", [["A", "B", "C"], ["1", "B", "=5+7"]])],
            None,
            [{"set": "MySheet", "A": "1", "B": "B", "C": "=5+7"}],
        ),
        pytest.param(
            [
                ("MySheet", [["A", "B", "C"], ["1", "B", "=5+7"]]),
                (
                    "OtherSheet",
                    [["D", "E", "F"], ["D1", "E1", "F1"], ["D2", "E2", "F2"]],
                ),
            ],
            None,
            [
                {"set": "MySheet", "A": "1", "B": "B", "C": "=5+7"},
                {"set": "OtherSheet", "D": "D1", "E": "E1", "F": "F1"},
                {"set": "OtherSheet", "D": "D2", "E": "E2", "F": "F2"},
            ],
        ),
        pytest.param(
            [
                ("MySheet", [["A", "B", "C"], ["1", "B", "=5+7"]]),
                (
                    "OtherSheet",
                    [["D", "E", "F"], ["D1", "E1", "F1"], ["D2", "E2", "F2"]],
                ),
            ],
            {"OtherSheet"},
            [{"set": "MySheet", "A": "1", "B": "B", "C": "=5+7"}],
        ),
    ],
)
def test_rows_from_workbook(
    sheets_and_rows: List[Tuple[str, List[List[str]]]],
    skip_sheets: Optional[Set[str]],
    expected: List[Dict[str, str]],
) -> None:
    workbook = openpyxl.Workbook()
    for sheet, rows in sheets_and_rows:
        worksheet = workbook.create_sheet(title=sheet)
        for row in rows:
            worksheet.append(row)
    del workbook["Sheet"]
    assert list(xlsx.rows_for_workbook(workbook, skip_sheets=skip_sheets)) == expected


def test_read_from_file(oracle: Oracle, tmp_path: Path) -> None:
    xlsx_path = tmp_path / "infile.xlsx"
    serializer = xlsx.XlsxDialect()
    workbook = openpyxl.Workbook()
    sheet = workbook["Sheet"]
    sheet.title = "S00"
    sheet.append(["scryfall_id", "nonfoil", "foil"])
    sheet.append(["5d5f3f57-410f-4ee2-b93c-f5051a068828", 3, 7])
    workbook.save(xlsx_path)
    collection = serializer.read(xlsx_path, oracle)
    assert collection.counts == {
        UUID("5d5f3f57-410f-4ee2-b93c-f5051a068828"): {
            CountType.NONFOIL: 3,
            CountType.FOIL: 7,
        }
    }



"""pytest test configuration file."""

from pathlib import Path
from typing import Dict, Generator, List
from uuid import UUID

import msgspec
import pytest
import responses

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.scryfall import fetcher
from mtg_ssm.scryfall.models import ScryCard, ScryList, ScryMigration, ScrySet

TEST_DATA_DIR = Path(__file__).parent / "data"
SETS_DATA_FILE = TEST_DATA_DIR / "sets.json"
CARDS_DATA_FILE = TEST_DATA_DIR / "cards.json"
MIGRATIONS_DATA_FILE = TEST_DATA_DIR / "migrations.json"


@pytest.fixture(autouse=True)
def _fetcher_disable_cache() -> Generator[None, None, None]:
    """Patch fetcher cache dirs for testing."""
    with fetcher.SESSION.cache_disabled():
        yield


@pytest.fixture(autouse=True)
def requests_mock() -> Generator[responses.RequestsMock, None, None]:
    """Auto replace all requests with a mock."""
    with responses.RequestsMock() as rsps:
        yield rsps


@pytest.fixture(scope="session")
def cards_data() -> List[ScryCard]:
    """Fixture containing all test card data."""
    with CARDS_DATA_FILE.open("rb") as card_data_file:
        return msgspec.json.decode(card_data_file.read(), type=List[ScryCard])


@pytest.fixture(scope="session")
def sets_data() -> List[ScrySet]:
    """Fixture containing all test set data."""
    with SETS_DATA_FILE.open("rb") as sets_data_file:
        sets_list = msgspec.json.decode(sets_data_file.read(), type=ScryList[ScrySet])
    return sets_list.data


@pytest.fixture(scope="session")
def migrations_data() -> List[ScryMigration]:
    """Fixture containing all test migrations data."""
    with MIGRATIONS_DATA_FILE.open("rb") as migrations_data_file:
        migrations_list = msgspec.json.decode(
            migrations_data_file.read(), type=ScryList[ScryMigration]
        )
    return migrations_list.data


@pytest.fixture(scope="session")
def id_to_card(cards_data: List[ScryCard]) -> Dict[UUID, ScryCard]:
    """Fixture returning scryfall id to card object for all test card data."""
    return {card.id: card for card in cards_data}


@pytest.fixture(scope="session")
def scryfall_data(
    cards_data: List[ScryCard],
    sets_data: List[ScrySet],
    migrations_data: List[ScryMigration],
) -> ScryfallDataSet:
    """Fixture containing all scryfall test data."""
    return ScryfallDataSet(sets=sets_data, cards=cards_data, migrations=migrations_data)



"""Card and set index container."""

import collections
import string
from typing import Dict, Iterable, List, Optional, Set, Tuple
from uuid import UUID

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.mtg import util
from mtg_ssm.scryfall.models import ScryCard, ScryMigrationStrategy, ScrySet


def name_card_sort_key(card: ScryCard) -> Tuple[str, int, str]:
    """Key function for sorting cards in a by-name list."""
    card_num, card_var = util.collector_int_var(card)
    return (card.set, card_num or 0, card_var or "")  # TODO: sort by set release date


def set_card_sort_key(card: ScryCard) -> Tuple[int, str]:
    """Key function for sorting cards in a by-set list."""
    card_num, card_var = util.collector_int_var(card)
    return (card_num or 0, card_var or "")


def build_snnmas(
    card: ScryCard,
) -> Iterable[Tuple[Optional[str], str, Optional[str], Optional[int], Optional[str]]]:
    """Build set, name, number, multiverse id tuple keys."""
    names_cnums: Set[Tuple[str, Optional[str]]] = {(card.name, card.collector_number)}
    for i, card_face in enumerate(card.card_faces or ()):
        names_cnums |= {
            (card_face.name, card.collector_number),
            (card_face.name, card.collector_number + string.ascii_lowercase[i]),
        }
    names_cnums |= {(n, None) for n, _ in names_cnums}

    sets: Set[Optional[str]] = {card.set, None}

    mvids: Set[Optional[int]] = {None} | set(card.multiverse_ids or ())

    artists: Set[Optional[str]] = {card.artist, None}

    for name, number in names_cnums:
        for set_ in sets:
            for mvid in mvids:
                for artist in artists:
                    yield (set_, name, number, mvid, artist)


class ScryfallDataIndex:
    """Card and set indexes for scryfall data."""

    def __init__(self) -> None:
        self.id_to_card: Dict[UUID, ScryCard] = {}
        self.name_to_cards: Dict[str, List[ScryCard]] = {}
        self.setcode_to_cards: Dict[str, List[ScryCard]] = {}
        self.id_to_setindex: Dict[UUID, int] = {}
        self.setcode_to_set: Dict[str, ScrySet] = {}
        self.migrate_old_id_to_new_id: Dict[UUID, UUID] = {}
        # snnma = Set, Name, (Collector) Number, Multiverse ID, Artist
        # TODO: convert to intersecting bitmap indexes
        # TODO: do we really need artist?
        self.snnma_to_id: Dict[
            Tuple[Optional[str], str, Optional[str], Optional[int], Optional[str]],
            Set[UUID],
        ] = {}

    def load_data(self, scrydata: ScryfallDataSet) -> None:
        """Load all cards and sets from a Scryfall data set."""
        self.id_to_card = {}
        self.id_to_setindex = {}
        self.setcode_to_set = {}

        self.snnma_to_id = collections.defaultdict(set)

        name_to_unsorted_cards: Dict[str, List[ScryCard]] = collections.defaultdict(list)
        setcode_to_unsorted_cards: Dict[str, List[ScryCard]] = {}

        for set_ in scrydata.sets:
            self.setcode_to_set[set_.code] = set_
            setcode_to_unsorted_cards[set_.code] = []

        for card in scrydata.cards:
            self.id_to_card[card.id] = card
            name_to_unsorted_cards[card.name].append(card)
            setcode_to_unsorted_cards[card.set].append(card)
            if not self.setcode_to_set[card.set].digital:
                for snnma in build_snnmas(card):
                    self.snnma_to_id[snnma].add(card.id)
        self.snnma_to_id = dict(self.snnma_to_id)

        for cards_list in name_to_unsorted_cards.values():
            cards_list.sort(key=name_card_sort_key)
        self.name_to_cards = dict(name_to_unsorted_cards)

        for cards_list in setcode_to_unsorted_cards.values():
            cards_list.sort(key=set_card_sort_key)
            self.id_to_setindex.update({c.id: i for i, c in enumerate(cards_list)})
        self.setcode_to_cards = dict(setcode_to_unsorted_cards)

        for migration in scrydata.migrations:
            if (
                migration.migration_strategy == ScryMigrationStrategy.MERGE
                and migration.new_scryfall_id
            ):
                self.migrate_old_id_to_new_id[
                    migration.old_scryfall_id
                ] = migration.new_scryfall_id


class Oracle:
    """Container for an indexed Scryfall data set."""

    def __init__(self, scrydata: ScryfallDataSet) -> None:
        self._scrydata = scrydata
        self.cards = scrydata.cards
        self.sets = scrydata.sets
        self.index = ScryfallDataIndex()
        self.index.load_data(scrydata)



#!/usr/bin/env python3
"""Script to build mtgjson testdata."""

import copy
from pathlib import Path

import msgspec

from mtg_ssm.scryfall import fetcher, models


class Error(Exception):
    """Base class for exceptions in this module."""


class MissingScryfallDataError(Error):
    """Exception raised for missing Scryfall data."""


TEST_DATA_DIR = Path(__file__).parent / "data"
TARGET_SETS_FILE = TEST_DATA_DIR / "sets.json"
TARGET_SETS_FILE1 = TEST_DATA_DIR / "sets1.json"
TARGET_SETS_FILE2 = TEST_DATA_DIR / "sets2.json"
TARGET_BULK_FILE = TEST_DATA_DIR / "bulk_data.json"
TARGET_CARDS_FILE = TEST_DATA_DIR / "cards.json"
TARGET_MIGRATIONS_FILE = TEST_DATA_DIR / "migrations.json"

SETS_NEXTPAGE_URL = "https://api.scryfall.com/sets?page=2"

TEST_SETS_TO_CARDS = {
    "arc": {"Leonin Abunas"},
    "bok": {"Faithful Squire // Kaiso, Memory of Loyalty"},
    "chk": {"Bushi Tenderfoot // Kenzo the Hardhearted", "Boseiju, Who Shelters All"},
    "csp": {"Jötun Grunt"},
    "fem": {"Thallid"},
    "hml": {"Cemetery Gate"},
    "hop": {"Akroma's Vengeance", "Dark Ritual"},
    "ice": {"Dark Ritual", "Forest", "Snow-Covered Forest"},
    "isd": {"Abattoir Ghoul", "Delver of Secrets // Insectile Aberration", "Forest"},
    "khm": {"Cosmos Elixir", "A-Cosmos Elixir"},
    "lea": {"Air Elemental", "Dark Ritual", "Forest"},
    "mbs": {"Hero of Bladehold", "Black Sun's Zenith"},
    "mma": {"Thallid"},
    "neo": {"Boseiju, Who Endures"},
    "oarc": {"All in Good Time"},
    "ogw": {"Wastes"},
    "ohop": {"Academy at Tolaria West"},
    "opc2": {"Akoum", "Chaotic Aether"},
    "p03": {"Goblin"},
    "pc2": {"Armored Griffin"},
    "pfrf": {"Dragonscale General"},
    "phop": {"Stairs to Infinity"},
    "phpr": {"Arena"},
    "plc": {"Boom // Bust"},
    "pls": {"Ertai, the Corrupted"},
    "pdci": {"Black Sun's Zenith", "Tazeem"},
    "pmbs": {"Hero of Bladehold"},
    "pneo": {"Boseiju, Who Endures"},
    "prna": {"Tithe Taker"},
    "prw2": {"Plains"},
    "ptg": {"Nightmare Moon // Princess Luna"},
    "ren": {"Sylvan Library"},
    "rna": {"Tithe Taker"},
    "s00": {"Rhox"},
    "sld": {"Goblin", "Viscera Seer", "Doubling Cube // Doubling Cube"},
    "sok": {"Erayo, Soratami Ascendant // Erayo's Essence"},
    "sunf": {"Happy Dead Squirrel"},
    "thb": {
        "Wolfwillow Haven",
        "Temple of Plenty",
        "Nyxbloom Ancient",
        "Klothys, God of Destiny",
    },
    "unh": {"Who // What // When // Where // Why"},
    "ust": {"Very Cryptic Command"},
    "vma": {"Academy Elite"},
    "w16": {"Serra Angel"},
}


def main() -> None:  # noqa: PLR0915
    """Read scryfall data and write a subset for use as test data."""
    print("Fetching scryfall data")
    scrydata = fetcher.scryfetch()
    bulk_data_list = msgspec.json.decode(
        fetcher._fetch_endpoint(fetcher.BULK_DATA_ENDPOINT),  # noqa: SLF001
        type=models.ScryList[models.ScryBulkData],
    )
    bulk_data = bulk_data_list.data

    print("Selecting sets")
    accepted_sets = sorted(
        (s for s in scrydata.sets if s.code in TEST_SETS_TO_CARDS),
        key=lambda sset: sset.code,
    )
    missing_sets = set(TEST_SETS_TO_CARDS.keys()) - {s.code for s in accepted_sets}
    if missing_sets:
        raise MissingScryfallDataError("Missing sets: " + str(missing_sets))

    print("Selecting cards")
    accepted_cards = sorted(
        (c for c in scrydata.cards if c.name in TEST_SETS_TO_CARDS.get(c.set, set())),
        key=lambda card: (card.set, card.name, card.collector_number, card.id),
    )
    accepted_card_ids = {c.id for c in accepted_cards}
    missing_cards = copy.deepcopy(TEST_SETS_TO_CARDS)
    for card in accepted_cards:
        missing_cards[card.set].discard(card.name)
    missing_cards = {k: v for k, v in missing_cards.items() if v}
    if missing_cards:
        raise MissingScryfallDataError("Missing cards: " + str(missing_cards))

    print("Selecting bulk data")
    accepted_bulk = sorted(
        (bd for bd in bulk_data if bd.type == fetcher.BULK_TYPE),
        key=lambda bulk: bulk.type,
    )

    print("Adjusting sets")
    for cset in accepted_sets:
        cset.card_count = len([c for c in accepted_cards if c.set == cset.code])

    print("Writing sets")
    sets_list = models.ScryList(
        data=accepted_sets,
        has_more=False,
        next_page=None,
        total_cards=None,
        warnings=None,
    )
    sets_list1 = models.ScryList(
        data=accepted_sets[: len(accepted_sets) // 2],
        has_more=True,
        next_page=SETS_NEXTPAGE_URL,
        total_cards=None,
        warnings=None,
    )
    sets_list2 = models.ScryList(
        data=accepted_sets[len(accepted_sets) // 2 :],
        has_more=False,
        next_page=None,
        total_cards=None,
        warnings=None,
    )
    TEST_DATA_DIR.mkdir(exist_ok=True)
    with TARGET_SETS_FILE.open("wb") as sets_file:
        sets_file.write(msgspec.json.format(msgspec.json.encode(sets_list), indent=2))
        sets_file.write(b"\n")
    with TARGET_SETS_FILE1.open("wb") as sets_file1:
        sets_file1.write(msgspec.json.format(msgspec.json.encode(sets_list1), indent=2))
        sets_file1.write(b"\n")
    with TARGET_SETS_FILE2.open("wb") as sets_file2:
        sets_file2.write(msgspec.json.format(msgspec.json.encode(sets_list2), indent=2))
        sets_file2.write(b"\n")

    print("Writing migrations")
    accepted_migrations = sorted(
        (m for m in scrydata.migrations if m.new_scryfall_id in accepted_card_ids),
        key=lambda migr: migr.id,
    )
    migrations_list = models.ScryList(
        data=accepted_migrations,
        has_more=False,
        next_page=None,
        total_cards=None,
        warnings=None,
    )
    with TARGET_MIGRATIONS_FILE.open("wb") as migrations_file:
        migrations_file.write(msgspec.json.format(msgspec.json.encode(migrations_list), indent=2))
        migrations_file.write(b"\n")

    print("Writing cards")
    with TARGET_CARDS_FILE.open("wb") as cards_file:
        cards_file.write(msgspec.json.format(msgspec.json.encode(accepted_cards), indent=2))
        cards_file.write(b"\n")

    print("Writing bulk data")
    bulk_list = models.ScryList(
        data=accepted_bulk,
        has_more=False,
        next_page=None,
        total_cards=None,
        warnings=None,
    )
    with TARGET_BULK_FILE.open("wb") as bulk_file:
        bulk_file.write(msgspec.json.format(msgspec.json.encode(bulk_list), indent=2))
        bulk_file.write(b"\n")

    print("Done")


if __name__ == "__main__":
    main()



"""Utility functions for working with card data."""

import functools
import string
from typing import Optional, Tuple

from mtg_ssm.scryfall.models import ScryCard

STRICT_BASICS = frozenset({"Plains", "Island", "Swamp", "Mountain", "Forest"})


@functools.lru_cache(maxsize=None)
def dig_str(collector_number: str) -> Tuple[Optional[int], Optional[str]]:
    """Split a collector number into integer portion and non-digit portion."""
    digpart = []
    strpart = []
    for char in collector_number:
        if char in string.digits:
            digpart.append(char)
        else:
            strpart.append(char)
    if not digpart:
        return (None, "".join(strpart))
    return (int("".join(digpart)), "".join(strpart) or None)


def collector_int_var(card: ScryCard) -> Tuple[Optional[int], Optional[str]]:
    """Get the integer and variant portions of a card's collector number."""
    return dig_str(card.collector_number)


def is_strict_basic(card_name: str) -> bool:
    """Is the card on of the five basic lands (not Snow or Wastes)."""
    return card_name in STRICT_BASICS



"""Tests for mtg_ssm.scryfall.fetcher."""

import re
from pathlib import Path
from typing import Dict, List, Pattern, Union

import pytest
from responses import RequestsMock

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.scryfall import fetcher
from mtg_ssm.scryfall.models import ScryCard, ScryMigration, ScrySet
from tests import gen_testdata

BULK_CARDS_REGEX = r"https://data\.scryfall\.io/default-cards/default-cards-\d{14}\.json"

ENDPOINT_TO_FILE: Dict[Union[str, Pattern[str]], Path] = {
    fetcher.BULK_DATA_ENDPOINT: gen_testdata.TARGET_BULK_FILE,
    fetcher.SETS_ENDPOINT: gen_testdata.TARGET_SETS_FILE1,
    gen_testdata.SETS_NEXTPAGE_URL: gen_testdata.TARGET_SETS_FILE2,
    fetcher.MIGRATIONS_ENDPOINT: gen_testdata.TARGET_MIGRATIONS_FILE,
    re.compile(BULK_CARDS_REGEX): gen_testdata.TARGET_CARDS_FILE,
}


@pytest.fixture()
def _scryurls(requests_mock: RequestsMock) -> None:
    """Populate mock responses for scryfall urls."""
    for endpoint, filename in ENDPOINT_TO_FILE.items():
        with filename.open("rb") as endpoint_file:
            requests_mock.add(
                "GET",
                endpoint,
                status=200,
                content_type="application/json",
                body=endpoint_file.read(),
                match_querystring=True,
            )


@pytest.mark.usefixtures("_scryurls")
def test_scryfetch() -> None:
    scrydata1 = fetcher.scryfetch()
    scrydata2 = fetcher.scryfetch()
    assert scrydata1 == scrydata2
    assert {s.code for s in scrydata1.sets} == gen_testdata.TEST_SETS_TO_CARDS.keys()
    assert {(c.set, c.name) for c in scrydata1.cards} == {
        (key, card) for key, value in gen_testdata.TEST_SETS_TO_CARDS.items() for card in value
    }


@pytest.mark.usefixtures("_scryurls")
def test_data_fixtures(
    scryfall_data: ScryfallDataSet,
    sets_data: List[ScrySet],
    cards_data: List[ScryCard],
    migrations_data: List[ScryMigration],
) -> None:
    scrydata = fetcher.scryfetch()
    assert scrydata == scryfall_data
    assert scrydata.sets == sets_data
    assert scrydata.cards == cards_data
    assert scrydata.migrations == migrations_data



"""Legacy record lookup capabilities for older file versions."""

from typing import Any, Dict, List, Optional, Set, Tuple
from uuid import UUID

from mtg_ssm.containers.indexes import Oracle
from mtg_ssm.mtg import util


class Error(Exception):
    """Base exception for this module."""


class NoMatchError(Error):
    """Raised when a matchor(Error):ing card cannot be found."""


class MultipleMatchError(Error):
    """Raised when multiple matching cards are found."""


COUNT_TYPE_TO_OLD_COUNT_TYPES: Dict[str, Set[str]] = {
    "nonfoil": {"nonfoil", "copies"},
    "foil": {"foil", "foils"},
}


def extract_counts(card_row: Dict[str, Any]) -> Dict[str, int]:
    """Convert old count type names to new count type names."""
    renamed_counts = {
        k: sum(int(card_row.get(t) or 0) for t in v)
        for k, v in COUNT_TYPE_TO_OLD_COUNT_TYPES.items()
    }
    return {k: v for k, v in renamed_counts.items() if v}


OTHER_SET_CODE_TO_SET_CODE = {
    # TODO: MOAR!
    "pMEI": ["phpr", "pdrc", "pbok", "pdtp", "pres", "purl", "pdp10", "pfrf", "pori"],
    "pPRE": ["psom", "pdgm", "pths", "pm15", "pktk", "pfrf", "pdtk", "pori", "pbfz"],
    "pJGP": [
        "jgp",
        "g99",
        "g00",
        "g01",
        "g02",
        "g03",
        "g04",
        "g05",
        "g06",
        "g07",
        "g08",
        "g09",
        "g11",
    ],
    "PO2": ["p02"],
    "pFNM": [
        "fnm",
        "f01",
        "f02",
        "f03",
        "f04",
        "f05",
        "f06",
        "f07",
        "f08",
        "f09",
        "f10",
        "f11",
        "f12",
        "f13",
        "f14",
        "f15",
        "f16",
        "f17",
        "f18",
    ],
    "NMS": ["nem"],
    "pARL": ["pal99", "pal00", "pal01", "pal02", "pal03", "pal04", "pal05", "pal06"],
    "VAN": ["pvan"],
    "pGTW": ["pg07", "pg08"],
    "pMGD": ["p10e", "pnph", "pori"],
    "pLPA": ["pmbs", "pavr", "pktk"],
    "pWPN": ["pwwk", "pwp10", "proe"],
    "HOP": ["ohop", "phop"],
    "PC2": ["opc2", "phop"],
    "pWCQ": ["ppro"],
    "DD3_JVC": ["jvc"],
    "DD3_GVL": ["gvl"],
    "DD3_EVG": ["evg"],
    "DD3_DVD": ["dvd"],
    "FRF_UGIN": ["ugin"],
    "PCA": ["opca"],
}

PSUDONYM_TO_ARTIST: Dict[Optional[str], str] = {
    "William Murai": "Willian Murai",
    "Dave Seeley": "David Seeley",
}


def find_scryfall_id(card_row: Dict[str, str], oracle: Oracle) -> UUID:
    """Heuristically determine the scryfall id for a given input row."""
    set_code = card_row.get("set", "")
    set_codes = [
        set_code,
        set_code.lower(),
        *OTHER_SET_CODE_TO_SET_CODE.get(set_code, []),
    ]
    name = card_row.get("name", "")
    collector_number = card_row.get("number") or None
    mvid = int(card_row.get("multiverseid") or -1)
    artist = card_row.get("artist") or None
    artist = PSUDONYM_TO_ARTIST.get(artist, artist)
    print(f"Searching => Set: {set_code}; Name: {name}; Number: {collector_number}; MVID: {mvid}")
    snnma_keys: List[Tuple[Optional[str], str, Optional[str], Optional[int], Optional[str]]] = []
    for set_ in set_codes:
        snnma_keys += [
            (set_, name, collector_number, None, None),
            (set_, name, None, mvid, None),
            (set_, name, None, None, artist),
            (set_, name, None, None, None),
            (None, name, None, None, artist),
        ]
    seen = False
    for snnma_key in snnma_keys:
        found = oracle.index.snnma_to_id.get(snnma_key)
        if found:
            seen = True
            scryfall_id = None
            if len(found) == 1:
                [scryfall_id] = found
            if util.is_strict_basic(name):
                scryfall_id = sorted(found)[0]
            if scryfall_id is not None:
                found_card = oracle.index.id_to_card[scryfall_id]
                print(
                    f"Found ==> Set: {found_card.set}; Name: {found_card.name}; Number: {found_card.collector_number}; MVIDs: {found_card.multiverse_ids}"
                )
                return scryfall_id
    if seen:
        msg = f"Could not find scryfall card for row: {card_row}"
        raise MultipleMatchError(msg)
    msg = f"Could not find scryfall card for row: {card_row}"
    raise NoMatchError(msg)


def coerce_row(card_row: Dict[str, Any], oracle: Oracle) -> Dict[str, Any]:
    """Coerce an unknown older row into a current row.

    If this cannot be done precisely, a warning will be displayed.
    """
    coerced_counts = extract_counts(card_row)
    if not coerced_counts:
        return {}
    coerced_scryfall_id = find_scryfall_id(card_row, oracle)
    return {"scryfall_id": coerced_scryfall_id, **coerced_counts}
